#include <glib.h>
#include <gtk/gtk.h>

#include <termios.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>

#include "xpiezo.h"
#include "xpiezo-gui.h"

gboolean
xpiezo_data_read (GIOChannel   *channel,
                  GIOCondition  condition,
                  gpointer      data)
{
  XPiezoContext *context = (XPiezoContext *) data;
  guint  bytes_read, i;
  guchar buffer[2048];

  g_io_channel_read_chars (channel, (gchar *) buffer, 2048, &bytes_read, NULL);

  g_printerr ("read:\n");
  for (i = 0; i < bytes_read; i++)
    g_printerr (i % 16 != 15 ? "%02x " : "%02x\n", (unsigned char) buffer[i]);

  if (i % 16 != 0)
    g_printerr ("\n");

  return TRUE;
}


gboolean
xpiezo_data_write_helper (GIOChannel   *channel,
                          GIOCondition  condition,
                          gpointer      data)
{
  XPiezoContext *context = (XPiezoContext *) data;
  gsize bytes_written;
  GError *error = NULL;

  g_io_channel_write_chars (channel,
                            context->write_buf + context->written,
                            context->write_n_bytes - context->written,
                            &bytes_written, &error);
  if (error)
    g_printerr ("write-error: %s\n", error->message);

  g_printerr ("write: %d bytes\n", bytes_written);

  context->written += bytes_written;

  if (context->written < context->write_n_bytes)
    {
      return TRUE;
    }
  else
    {
      g_free (context->write_buf);
      context->write_buf = NULL;
      context->write_n_bytes = 0;
      context->written = 0;
      return FALSE;
    }
}


gboolean
xpiezo_data_write (XPiezoContext *context,
                   gchar         *data,
                   gint           n_bytes)
{
  if (context->written < context->write_n_bytes)
    return FALSE;

  if (context->write_buf)
    g_free (context->write_buf);

  context->write_buf = g_memdup (data, n_bytes);
  context->write_n_bytes = n_bytes;
  context->written = 0;

  g_io_add_watch (context->serial, G_IO_OUT,
                  xpiezo_data_write_helper, context);

  return TRUE;
}


GIOChannel *
xpiezo_open_piezo (XPiezoContext *context,
                   gchar         *serial_name,
                   GError       **error)
{
  GError *tmp_error = NULL;
  struct termios tio;

  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  if (context->serial_name)
    g_free (context->serial_name);

  context->serial_name = g_strdup (serial_name);
  context->serial = g_io_channel_new_file (serial_name, "r+", &tmp_error);

  if (tmp_error != NULL)
    {
      g_propagate_error (error, tmp_error);
      return NULL;
    }

  context->serial_fd = g_io_channel_unix_get_fd (context->serial);

  if (tcgetattr (context->serial_fd, &tio) != 0)
    {
      g_io_channel_unref (context->serial);
      tmp_error = g_error_new (G_IO_CHANNEL_ERROR,
                               g_io_channel_error_from_errno (errno),
                               "%s",
                               g_strerror (errno));
      g_propagate_error (error, tmp_error);
      return NULL;
    }

  cfmakeraw (&tio);
  if (cfsetspeed (&tio, B38400) != 0)
    {
      g_io_channel_unref (context->serial);
      tmp_error = g_error_new (G_IO_CHANNEL_ERROR,
                               g_io_channel_error_from_errno (errno),
                               "%s",
                               g_strerror (errno));
      g_propagate_error (error, tmp_error);
      return NULL;
    }

  if (tcsetattr (context->serial_fd, TCSANOW, &tio) != 0)
    {
      g_io_channel_unref (context->serial);
      tmp_error = g_error_new (G_IO_CHANNEL_ERROR,
                               g_io_channel_error_from_errno (errno),
                               "%s",
                               g_strerror (errno));
      g_propagate_error (error, tmp_error);
      return NULL;
    }

  g_io_channel_set_encoding (context->serial, NULL, NULL);
  g_io_channel_set_buffered (context->serial, FALSE);
  g_io_add_watch (context->serial, G_IO_IN, xpiezo_data_read, context);

  return context->serial;
}


gboolean
xpiezo_sensors_request (gpointer data)
{
  XPiezoContext *context = (XPiezoContext *) data;

  xpiezo_data_write (context, "UFOPID", 6);

  return TRUE;
}


XPiezoPreset
xpiezo_preset_from_pst (gchar *pst, gint n_bytes)
{
  XPiezoPreset preset = { 0, };

  g_return_val_if_fail (n_bytes == 56, preset);

#define GUINT32_FROM_LE_BYTES(x)  ((guchar) (x)[0])        + \
                                 (((guchar) (x)[1]) <<  8) + \
                                 (((guchar) (x)[2]) << 16) + \
                                 (((guchar) (x)[3]) << 24);

  preset.nr_kp          = GUINT32_FROM_LE_BYTES (pst +  8);
  preset.nr_kd          = GUINT32_FROM_LE_BYTES (pst + 12);
  preset.yaw_kp         = GUINT32_FROM_LE_BYTES (pst + 16);
  preset.yaw_kd         = GUINT32_FROM_LE_BYTES (pst + 20);
  preset.nr_k_stick     = GUINT32_FROM_LE_BYTES (pst + 24);
  preset.yaw_k_stick    = GUINT32_FROM_LE_BYTES (pst + 28);
  preset.nr_k_expo      = GUINT32_FROM_LE_BYTES (pst + 32);
  preset.angle_max      = GUINT32_FROM_LE_BYTES (pst + 36);
  preset.nr_k_expo_edge = GUINT32_FROM_LE_BYTES (pst + 40);
  preset.k_fast_yaw     = GUINT32_FROM_LE_BYTES (pst + 44);

#undef GUINT32_FROM_LE_BYTES
  return preset;
}

void
xpiezo_preset_save_to_file (XPiezoPreset  *preset,
                            gchar         *filename,
                            GError       **error)
{
  GKeyFile *kfile;
  gchar    *section = "x3d preset";
  gchar    *data;

  kfile = g_key_file_new ();
  g_key_file_load_from_data (kfile, "# Parameter set for the X-3D", 28,
                             G_KEY_FILE_KEEP_COMMENTS |
                             G_KEY_FILE_KEEP_TRANSLATIONS,
                             NULL);

#define SET_INT(field) g_key_file_set_integer (kfile, section,\
                                               #field, (gint) preset->field)
#define SET_BOOL(field) g_key_file_set_boolean (kfile, section,\
                                               #field, preset->field)
  SET_BOOL (easyparams);
  SET_INT (orientation);
  SET_INT (weight);
  SET_INT (agility);
  SET_INT (temp_comp);

  SET_INT (nr_kp);
  SET_INT (nr_kd);
  SET_INT (angle_max);
  SET_INT (nr_k_stick);
  SET_INT (nr_k_expo);
  SET_INT (nr_k_expo_edge);
  SET_INT (nr_k_stick_direct);

  SET_INT (yaw_kp);
  SET_INT (yaw_kd);
  SET_INT (yaw_k_stick);
  SET_BOOL (fast_yaw);
  SET_INT (fast_yaw_margin);
  SET_INT (k_fast_yaw);
#undef SET_INT
#undef SET_BOOL

  g_key_file_set_comment (kfile, section, "nr_kp",
                          "\n nick/roll parameters\n", NULL);
  g_key_file_set_comment (kfile, section, "yaw_kp",
                          "\n yaw parameters\n", NULL);
  data = g_key_file_to_data (kfile, NULL, NULL);
  g_printerr ("%s", data);
  g_free (data);

  g_key_file_free (kfile);
}


int
main (int argc, char *argv[])
{
  XPiezoContext *context;
  GtkWidget *window;
  GError *error = NULL;

  XPiezoPreset preset = { TRUE,
                          XPIEZO_ORI_000,
                          6,
                          0,
                          TRUE,
                          220, 470, 1000000, 8, 60, 1000, 10,
                          410, 535, 20, 15, 0,
                          60 };
  char *pst = "p\x14I\x8a\x03\x00\x00\x00\xdc\x00\x00\x00"
              "\xd6\x01\x00\x00\x9a\x01\x00\x00\x17\x02\x00\x00"
              "\x08\x00\x00\x00\x14\x00\x00\x00<\x00\x00\x00"
              "@B\x0f\x00\xe8\x03\x00\x00<\x00\x00\x00"
              "\x00\x0f2\n\xc2\x04\x84\x00";

  gtk_set_locale ();
  gtk_init (&argc, &argv);

  g_printerr ("config: %s\n",g_get_user_config_dir ());
  g_printerr ("data:   %s\n",g_get_user_config_dir ());

  context = g_new0 (XPiezoContext, 1);

  xpiezo_load_pixmaps (context);

  window = xpiezo_create_window (context);

  gtk_widget_show (window);

  preset = xpiezo_preset_from_pst (pst, 56);
  xpiezo_preset_save_to_file (&preset, NULL, NULL);
  xpiezo_open_piezo (context, "/dev/ttyUSB0", &error);

  if (error)
    g_printerr ("error opening: %s\n", error->message);

  xpiezo_data_write (context, "UFOPI\02", 6);

  g_timeout_add (1500, xpiezo_sensors_request , context);
  gtk_main ();
  return 0;
}
